//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
extern "C" {
#include <ggl.h>
}

#include "memblock.h"
#include "gstring.h"
#include "gui.h"
#include "resources.h"


// CLASS gApplication IMPLEMENTATION

gApplication::gApplication() : gControl()
{
	// INITIALIZE GUI APPLICATION

	contrastlevel=7;
	
	// INITIALIZE SCREEN
	int size=(SCREEN_W>>1)*SCREEN_H+16;
	screenbuffer=(unsigned int *)sys_phys_malloc(size);
	if(!screenbuffer) InsufMemoryError();
	// CLEAR_SCREEN
	memset((char *)sys_map_v2p((unsigned int)screenbuffer),0,size);
	
	drawsurf.addr=(int *) ((sys_map_v2p((unsigned int)screenbuffer)+15)&0xfffffff0);
	drawsurf.width=LCD_W;
	drawsurf.x=drawsurf.y=0;	
	sizex=131;
	sizey=80;
	
	ClearClipRegion();
	
	
	// SHOW FRAME BUFFER
	lcd_setmode(MODE_16GRAY,drawsurf.addr);
	
	// SET PROPER CONTRAST
	// TODO: SAVE/RESTORE USER SETTINGS ON CONTRAST
	lcd_setcontrast(contrastlevel);
	

	// TODO: INITIALIZE OTHER HARDWARE
	
	// TODO: INITIALIZE LOOP PARAMETERS
	
	RunLoopFlags=0;
	RunResult=0;
	CurrentForm=NULL;
}

gApplication::~gApplication()
{
	// CLEANUP PROCEDURE
	
	// TODO: CLOSE OTHER HARDWARE FIRST
	
	
	// TODO: RESTORE CONTRAST
	
	// NO NEED TO CHANGE VIDEO MODE: WILL BE RESTORED AUTOMATICALLY ON EXIT
	
	// FREE MEMORY

	free(screenbuffer);
	screenbuffer=NULL;
	drawsurf.addr=NULL;
	
}


void gApplication::Update(gUpdate *u)
{

	gUpdate uclient;
	
	uclient.clipx=u->clipx-posx+viewx;
	uclient.clipy=u->clipy-posy+viewy;
	uclient.clipx2=u->clipx2-posx+viewx;
	uclient.clipy2=u->clipy2-posy+viewy;

	// PASS UPDATE MESSAGE TO CHILDS
	SetClipRegion(uclient.clipx,uclient.clipy,uclient.clipx2,uclient.clipy2);
	
	gControl *chld=(UpdateFirst? UpdateFirst:FirstChild);
	
	while(chld) {
		chld->Update(&uclient);
		chld=chld->Next;
	}
	
	invx=invy=invx2=invy2=-1;		// MARK UPDATED
	ClearClipRegion();
	UpdateFirst=NULL;
	// DONE, ALL CHILDS UPDATED
}

void gApplication::Update()
{

	if(invx<0) return;
	
	gUpdate uclient;
	
	uclient.clipx=invx;
	uclient.clipy=invy;
	uclient.clipx2=invx2;
	uclient.clipy2=invy2;

	Update(&uclient);
}



int gApplication::Run(gForm *Main)
{

	Invalidate();
	Update();

	// START EVENT LOOP
	gEvent e;
	gControl *Idleptr;
	
	RunLoopFlags=0;
	RunResult=0;
	

	// SEND CREATE EVENT TO ALL FORMS
	CurrentForm=FirstChild;

	// PERFORM APPLICATION START SEQUENCE, SENDING SEVERAL EVENTS
	if(!CurrentForm) return 0;

	
	while(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMCREATE;
	e.Data=NULL;
	CurrentForm->DoEvent(&e);
	CurrentForm=CurrentForm->Next;
	}

	if(Main) CurrentForm=Main;
	else CurrentForm=LastChild;
	
	
	// SEND EVENT OPEN ONLY TO THE CURRENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMOPEN;
	e.Data=NULL;
	CurrentForm->DoEvent(&e);

	// SEND EVENT GETFOCUS ONLY TO THE CURRENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_GETFOCUS;
	e.Data=NULL;
	CurrentForm->DoEvent(&e);

	
	unsigned int a;
	
	Idleptr=CurrentForm;
	
	while(!(RunLoopFlags&GAPP_EXIT)) {

		if(!CurrentForm) break;		// NO MORE FORMS, EXIT
		
		// CHECK FOR KEYBOARD EVENTS
		
		a=keyb_getmsg();
		if(a) {
		e.Type=GEVENT_KEYB;
		//  THE GUI USES ITS OWN ALPHA MODE, SO THE ALPHA SHIFT PLANE MUST BE FILTERED OUT
		e.Arg=a&~SHIFT_ALPHA;
		
		e.Data=NULL;

		if(!CurrentForm->DoEvent(&e))
		{
			DoEventExclusive(&e);	// IF THE FORM DID NOT HANDLE THE EVENT, PASS IT TO THE APPLICATION
		}
		
		Update();
		continue;			
		}
		
		
		// CHECK FOR OTHER HARDWARE EVENTS
		// TODO
		
		
		
		// PERFORM IDLE EVENT
		e.Type=GEVENT_IDLE;
		e.Arg=0;
		e.Data=NULL;
		if(!Idleptr->DoEvent(&e)) {
			DoEventExclusive(&e);	// SEND IDLE EVENT TO APPLICATION
		}
		Idleptr=Idleptr->Prev;
		if(!Idleptr) Idleptr=LastChild;		// CYCLE THROUGH ALL FORMS
		
		// PERFORM SCREEN UPDATE
		Update();
		
	}
	
	// SEND EVENT LOSEFOCUS ONLY TO THE CURRENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_LOSEFOCUS;
	e.Data=NULL;
	if(!CurrentForm->DoEvent(&e)) DoEventExclusive(&e);
	
	// SEND EVENT OPEN ONLY TO THE CURRENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMCLOSE;
	e.Data=NULL;
	if(!CurrentForm->DoEvent(&e)) DoEventExclusive(&e);
	
	// SEND DESTROY EVENT TO ALL FORMS
	CurrentForm=FirstChild;
	
	while(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMEXIT;
	e.Data=NULL;
	if(!CurrentForm->DoEvent(&e)) DoEventExclusive(&e);
	CurrentForm=CurrentForm->Next;
	}

	
	return RunResult;
}

void gApplication::Terminate(int result)
{
	RunResult=result;
	RunLoopFlags|=GAPP_EXIT;
}

BOOL gApplication::QueryTerminate()
{
	// SEND QUERY EVENT TO ALL FORMS, RETURN TRUE IF ALL FORMS ARE OK TO TERMINATE
	// RETURNS FALSE IF APPLICATION NEEDS TO KEEP RUNNING.
	gEvent e;
	
	gControl *cform;
	cform=FirstChild;

	// SEND FIRST TO CURRENT FORM, THEN THE REST
	if(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_QUERYCLOSE;
	e.Data=NULL;
	if(!CurrentForm->DoEvent(&e)) return FALSE;
	}
	
	while(cform) {
		if(cform!=CurrentForm) {
			e.Type=GEVENT_GUI;
			e.Arg=GEVGUI_QUERYCLOSE;
			e.Data=NULL;
			if(!cform->DoEvent(&e)) return FALSE;
		}
	cform=cform->Next;
	}
	
	return TRUE;
}

void gApplication::Invalidate(gControl *Sender)
{
	if(Sender) {
	if(UpdateFirst) {
		gControl *ch=Sender;
		while(ch) { if(UpdateFirst==ch) { UpdateFirst=Sender; break; } ch=ch->Next; }
	}
	else UpdateFirst=Sender;
	} else UpdateFirst=NULL;
	
	invx=posx;
	invy=posy;
	invx2=posx+sizex-1;
	invy2=posy+sizey-1;
}

void gApplication::Invalidate(gUpdate *_u,gControl *Sender)
{
	if(Sender) {
	if(UpdateFirst) {
		gControl *ch=Sender;
		while(ch) { if(UpdateFirst==ch) { UpdateFirst=Sender; break; } ch=ch->Next; }
	}
	else UpdateFirst=Sender;
	} else UpdateFirst=NULL;

	if(_u->clipx>_u->clipx2) {
		int tmp=_u->clipx;
		_u->clipx=_u->clipx2;
		_u->clipx2=tmp;
		
	}
	if(_u->clipy>_u->clipy2) {
		int tmp=_u->clipy;
		_u->clipy=_u->clipy2;
		_u->clipy2=tmp;
	}
	
	if(invx<0) {
		invx=_u->clipx;
		invy=_u->clipy;
		invx2=_u->clipx2;
		invy2=_u->clipy2;
	}
	else {
	invx=min(_u->clipx,invx);
	invy=min(_u->clipy,invy);
	invx2=max(_u->clipx2,invx2);
	invy2=max(_u->clipy2,invy2);
	}
	
	if(invx<posx) invx=posx;
	if(invy<posy) invy=posy;
	if(invx2>=sizex+posx) invx2=posx+sizex-1;
	if(invy2>=sizey+posy) invy2=posy+sizey-1;
	
}

void gApplication::OpenNewForm(gControl *ptr)
{
	gEvent e;
	
	// CREATE NEW FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMCREATE;
	e.Data=0;
	ptr->DoEvent(&e);
	
	if(CurrentForm) {
	// TAKE FOCUS AWAY FROM CURENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_LOSEFOCUS;
	e.Data=0;
	CurrentForm->DoEvent(&e);
	}
	
	// OPEN NEW FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMOPEN;
	e.Data=0;
	ptr->DoEvent(&e);
	
	// SET FOCUS
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_GETFOCUS;
	e.Data=0;
	ptr->DoEvent(&e);
	
	CurrentForm=ptr;
	Invalidate();
}

void gApplication::CloseForm(gControl *ptr)
{
	gEvent e;
	if(ptr==CurrentForm) {
	// TAKE FOCUS AWAY FROM CURENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_LOSEFOCUS;
	e.Data=0;
	CurrentForm->DoEvent(&e);

	}
	
	// CLOSE FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMCLOSE;
	e.Data=0;
	ptr->DoEvent(&e);
	

	if(ptr==CurrentForm) {
	// SET FOCUS
	CurrentForm=CurrentForm->Prev;
	if(!CurrentForm && (ptr!=LastChild)) CurrentForm=LastChild;

	if(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_GETFOCUS;
	e.Data=0;
	CurrentForm->DoEvent(&e);
	}
	}
	
}

void gApplication::SetCurrentForm(gControl *ptr)
{
	gEvent e;
	if(ptr==CurrentForm) return;
	
	// TAKE FOCUS AWAY FROM CURENT FORM
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_LOSEFOCUS;
	e.Data=0;
	CurrentForm->DoEvent(&e);

	CurrentForm=ptr;
	// SET FOCUS

	if(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_GETFOCUS;
	e.Data=0;
	CurrentForm->DoEvent(&e);
	
}
}


int gApplication::RunInternalLoop()
{
	int BackupFlags=RunLoopFlags;
	int BackupResult=RunResult;
	// START INTERNAL EVENT LOOP
	gEvent e;
	gControl *Idleptr;
	unsigned int a;
	
	Idleptr=CurrentForm;
	RunLoopFlags=0;
	
	while(!(RunLoopFlags&GAPP_EXIT)) {

		if(!CurrentForm) break;		// NO MORE FORMS, EXIT
		
		// CHECK FOR KEYBOARD EVENTS
		
		a=keyb_getmsg();
		if(a) {
		e.Type=GEVENT_KEYB;
		//  THE GUI USES ITS OWN ALPHA MODE, SO THE ALPHA SHIFT PLANE MUST BE FILTERED OUT
		e.Arg=a&~SHIFT_ALPHA;
		
		e.Data=NULL;

		if(!CurrentForm->DoEvent(&e))
		{
			DoEventExclusive(&e);	// IF THE FORM DID NOT HANDLE THE EVENT, PASS IT TO THE APPLICATION
		}
		
		Update();
		continue;			
		}
		
		
		// CHECK FOR OTHER HARDWARE EVENTS
		// TODO
		
		
		
		// PERFORM IDLE EVENT
		e.Type=GEVENT_IDLE;
		e.Arg=0;
		e.Data=NULL;
		if(!Idleptr->DoEvent(&e)) {
			DoEventExclusive(&e);	// SEND IDLE EVENT TO APPLICATION
		}
		Idleptr=Idleptr->Prev;
		if(!Idleptr) Idleptr=LastChild;		// CYCLE THROUGH ALL FORMS
		
		// PERFORM SCREEN UPDATE
		Update();
		
	}
	
	CloseForm(CurrentForm);

	RunLoopFlags=BackupFlags;
	int tmp=RunResult;
	RunResult=BackupResult;
	return tmp;
}

gLoopData * gApplication::StartOpenLoop(gForm *main)
{
	
 gLoopData *v=new gLoopData;
 if(!v) { InsufMemoryError(); return NULL; }
 v->BackupFlags=RunLoopFlags;
 v->BackupResult=RunResult;
 if(!main) main=(gForm *)LastChild;
 v->Idleptr=main;
 RunLoopFlags=0;
 RunResult=0;
 if(main) OpenNewForm(main);
 return v;
}

BOOL gApplication::RunOpenLoopOnce(gLoopData *v)
{
	gEvent e;
	int a;
	if((RunLoopFlags&GAPP_EXIT)) return FALSE;

		if(!CurrentForm) return FALSE;		// NO MORE FORMS, EXIT
		
		// CHECK FOR KEYBOARD EVENTS
		
		a=keyb_getmsg();
		while(a) {
		e.Type=GEVENT_KEYB;
		//  THE GUI USES ITS OWN ALPHA MODE, SO THE ALPHA SHIFT PLANE MUST BE FILTERED OUT
		e.Arg=a&~SHIFT_ALPHA;
		
		e.Data=NULL;

		if(!CurrentForm->DoEvent(&e))
		{
			DoEventExclusive(&e);	// IF THE FORM DID NOT HANDLE THE EVENT, PASS IT TO THE APPLICATION
		}
		
		Update();
		a=keyb_getmsg();
		}
		
		
		
		// PERFORM IDLE EVENT
		e.Type=GEVENT_IDLE;
		e.Arg=0;
		e.Data=NULL;
		if(!v->Idleptr->DoEvent(&e)) {
			DoEventExclusive(&e);	// SEND IDLE EVENT TO APPLICATION
		}
		v->Idleptr=v->Idleptr->Prev;
		if(!v->Idleptr) v->Idleptr=LastChild;		// CYCLE THROUGH ALL FORMS
		
		// PERFORM SCREEN UPDATE
		Update();
		

		return TRUE;
		
}

int gApplication::EndOpenLoop(gLoopData *v)
{
	if(CurrentForm) CloseForm(CurrentForm);

	RunLoopFlags=v->BackupFlags;
	int tmp=RunResult;
	RunResult=v->BackupResult;
	return tmp;
	delete v;
}

void gApplication::CloseAllAndExit()
{
	while(CurrentForm)	CloseForm(CurrentForm);

	// SEND DESTROY EVENT TO ALL FORMS
	CurrentForm=FirstChild;
	
	gEvent e;
	
	while(CurrentForm) {
	e.Type=GEVENT_GUI;
	e.Arg=GEVGUI_FORMEXIT;
	e.Data=NULL;
	if(!CurrentForm->DoEvent(&e)) DoEventExclusive(&e);
	CurrentForm=CurrentForm->Next;
	}
	return;
}
