//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
extern "C" {
#include <ggl.h>
}

#include "gui.h"
#include "resources.h"


gListBox::gListBox(gControl& Parent,int x,int y,int width,int height,int _Style,gFont *_Font)
: gControl(Parent)
{
	int w,h;

	memcpy(&Font,_Font,sizeof(gFont));
	
	ItemHeight=max(CHECKHEIGHT,Font.TextHeight());
	
	if(width<0) {
		// WIDTH SPECIFIED IN CHARACTERS
		w=-width*Font.MaxWidth;
		if(_Style&S_VSCROLLBAR) w+=SCROLLBARWIDTH;
		if(_Style&S_HASBORDER) w+=2*BORDERWIDTH+2;
	}
	else w=width;

	if(height<0) {
		// HEIGHT SPECIFIED IN LINES OF TEXT
		h=-height*ItemHeight;
		if(_Style&S_HSCROLLBAR) h+=SCROLLBARWIDTH;
		if(_Style&S_HASBORDER) h+=2*BORDERWIDTH+2;
	}
	else {
		int tmph=ItemHeight;
		if(_Style&S_HSCROLLBAR) tmph+=SCROLLBARWIDTH;
		if(_Style&S_HASBORDER) tmph+=2*BORDERWIDTH+2;
		if(tmph>height) h=tmph;
		else h=height;
	}
	
	StyleFlags=_Style;

	ncx=x;
	ncy=y;
	posx=x;
	posy=y;
	ncwidth=w;
	ncheight=h;

	// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
	
	if(StyleFlags & S_HASBORDER) { posx+=BORDERWIDTH+1; posy+=BORDERWIDTH+1; h-=2*BORDERWIDTH+2; w-=2*BORDERWIDTH+2; }
	if(StyleFlags&S_HSCROLLBAR) {
		h-=SCROLLBARWIDTH;
	}
	if(StyleFlags&S_VSCROLLBAR) {
		w-=SCROLLBARWIDTH;
	}
	
	DispItems=h/ItemHeight;
	if(DispItems<=0) DispItems=1;

	vieww=w;
	viewh=h;
	viewx=viewy=0;
	if(!Resize(w,h)) InsufMemoryError();

	ClearClipRegion();
	NumItems=0;	
	SelectedItem=0;
	
	Color=GBLACK;
	BkColor=GWHITE;
	SelColor=GGRAY4;
	ViewItem=0;
	MaxWidth=w;
	Items=(gListItem *)malloc(ITEM_GROUP*sizeof(gListItem));
	MaxItems=ITEM_GROUP;
	
	if(StyleFlags&S_HSCROLLBAR) scrollh=1;
	if(StyleFlags&S_VSCROLLBAR) scrollv=1;

}

gListBox::~gListBox()
{
		if(Items) delete Items;
}

BOOL gListBox::ReallocList(int _NumItems)
{
	int nblocks=((_NumItems+(ITEM_GROUP-1))/ITEM_GROUP);
	
	if(nblocks*ITEM_GROUP-MaxItems>2*ITEM_GROUP) {
		// RELEASE EXTRA MEMORY
		Items=(gListItem *)realloc(Items,nblocks*ITEM_GROUP*sizeof(gListItem));
		MaxItems=nblocks*ITEM_GROUP;
		return TRUE;
	}
	if(nblocks*ITEM_GROUP>MaxItems) {
		gListItem *temp=(gListItem *)realloc(Items,nblocks*ITEM_GROUP*sizeof(gListItem));
		if(!temp) return FALSE;
		MaxItems=nblocks*ITEM_GROUP;
		Items=temp;
	}
	return TRUE;
}



void gListBox::SetItemList(gListItem *List,int _NumItems)
{
	if(!ReallocList(_NumItems)) InsufMemoryError();
	memcpy(Items,List,sizeof(gListItem)*_NumItems);
	NumItems=_NumItems;
	ListChanged();
}

void gListBox::ListChanged()
{
	int redraw=0;
	int maxwidth,f,temp;
	if(ViewItem>=NumItems-DispItems) { ViewItem=NumItems-DispItems; redraw=1; }
	if(ViewItem<0) { ViewItem=0; redraw=1; }
	if(SelectedItem<0) { SelectedItem=0; redraw=1; }
	if(SelectedItem>=NumItems) { SelectedItem=NumItems-1; redraw=1; }

	maxwidth=0;
	for(f=0;f<NumItems;++f)
	{
		Items[f].StateFlags&=~LS_SELECTED;
		temp=CalcItemWidth(f);
		if(temp>maxwidth) maxwidth=temp;
	}
	
	if(SelectedItem<NumItems && (StyleFlags&S_FOCUSED)) Items[SelectedItem].StateFlags|=LS_SELECTED;
	
	if(maxwidth<vieww) maxwidth=vieww;
	if(maxwidth>MaxWidth) { redraw=1; }
	if(!Resize(maxwidth,sizey)) InsufMemoryError();
	
	
	MaxWidth=maxwidth;
	if(viewx+vieww>MaxWidth) { viewx=MaxWidth-vieww; redraw=1; }
	
	if(redraw) { ClearClipRegion(); RedrawAll(); Invalidate(); }
}


void gListBox::InsertItem(int Index,char *Text, int TreeLevel,int StateFlags,int Data)
{
	if(Index<0 || Index>NumItems) return;
	if(NumItems>=MaxItems) {
		// REALLOC
		if(!ReallocList(NumItems+1)) { InsufMemoryError(); }
	}
	
	memmove(Items+(Index+1),Items+Index,(NumItems-Index)*sizeof(gListItem));
	Items[Index].StateFlags=StateFlags;
	Items[Index].Text=Text;
	Items[Index].TreeLevel=TreeLevel;
	Items[Index].Data=Data;
	++NumItems;
}
void gListBox::RemoveItem(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	memmove(Items+Index,Items+Index+1,(NumItems-Index)*sizeof(gListItem));
	--NumItems;
	
}
void gListBox::AddItem(char *Text, int TreeLevel,int StateFlags,int Data)
{
	InsertItem(NumItems,Text,TreeLevel,StateFlags,Data);
}

gListItem *gListBox::GetItem(int Index)
{
	if(Index<0 || Index>NumItems) return NULL;
	return Items+Index;
}


void gListBox::RedrawAll()
{
	int f,nitem,y;
	for(f=0,nitem=ViewItem,y=0;f<DispItems;++f)
	{
		while(!DrawItem(nitem,y)) { ++nitem; }
		++nitem;
		y+=ItemHeight;
	}
}
int gListBox::CalcItemWidth(int Index)
{
	int width=Font.TextWidth(Items[Index].Text);
	width+=Items[Index].TreeLevel*LEVELWIDTH;
	if(Items[Index].StateFlags&LS_EXPANDABLE) width+=CHECKWIDTH;
	if(Items[Index].StateFlags&LS_HASCHECK) width+=CHECKWIDTH;
	++width;
	return width;
}

BOOL gListBox::DrawItem(int Index,int y)
{
	int x=0;
	int col=BkColor;
	gListItem *i=Items+Index;

	// ERASE BACKGROUND
	if(Index>=0 && Index<NumItems) {
		if(i->StateFlags&LS_SELECTED) col=SelColor;
		if((i->StateFlags&LS_HIDDEN)) return FALSE;	// ITEM IS NOT VISIBLE
	}
	Rect(x,y,x+MaxWidth-1,y+ItemHeight-1,col);

	if(Index>=NumItems) return TRUE;
	
	
	int level=1,nxtlevel;
	nxtlevel=NextVisibleItem(Index);
	if(nxtlevel>=0) nxtlevel=Items[nxtlevel].TreeLevel;
	else nxtlevel=0;

	while(level<=i->TreeLevel) {

		x+=LEVELWIDTH;	// MOVE TO THE RIGHT
		// DRAW TREE LINES
		if( ((Index<NumItems-1) && (nxtlevel<level)) || (Index==NumItems-1)) {
			// DRAW AS LAST ITEM
			VLine(x-5,y,y+ItemHeight/2,GBLACK);
			if(level<i->TreeLevel) HLine(y+ItemHeight/2,x-4,x,GBLACK);

		}
		else VLine(x-5,y,y+ItemHeight-1,GBLACK);
		if(level==i->TreeLevel) HLine(y+ItemHeight/2,x-4,x-2,GBLACK);
		
		++level;
	}
	
	
	if(i->StateFlags&LS_EXPANDABLE) {
		// DRAW COLLAPSE/EXPAND SIGN
		int type;
		if(i->StateFlags&LS_EXPANDED) type=S_MINUS;
		else type=S_PLUS;
		DrawIconPartial(x,y+(ItemHeight-CHECKHEIGHT+1)/2,iCheckBox,CHECKBMPWIDTH,type*CHECKWIDTH,0,CHECKWIDTH,CHECKHEIGHT);
		x+=CHECKWIDTH;
		}
		
	if(i->StateFlags&LS_HASCHECK) {
		// DRAW CHECKBOX
		int type;
		if(i->StateFlags&LS_CHECKED) type=S_CHECK;
		else type=S_UNCHECKED;
		if(i->StateFlags&LS_UNKNOWN) type=S_DOT;
		
		DrawIconPartial(x,y+(ItemHeight-CHECKHEIGHT+1)/2,iCheckBox,CHECKBMPWIDTH,type*CHECKWIDTH,0,CHECKWIDTH,CHECKHEIGHT);
		x+=CHECKWIDTH;
	}
	
	++x;
	if(i->StateFlags&LS_SELECTED) col=SelColor;

	// DRAW TEXT
	DrawText(x,y+(ItemHeight-Font.TextHeight()+1)/2,i->Text,&Font,Color);
	return TRUE;
}


void gListBox::Update(gUpdate *u)
{

	RedrawFrame();

	// UPDATE VIEWED PORTION
	
	drawsurf.x=(u->clipx<posx)? viewx:(u->clipx-posx+viewx);
	drawsurf.y=(u->clipy<posy)? viewy:(u->clipy-posy+viewy);
	
	if(drawsurf.x>=viewx+vieww) return;
	if(drawsurf.y>=viewy+viewh) return;
	
	int dw,dh;
	dw=u->clipx2-posx+viewx-drawsurf.x+1;
	if(dw<=0) return;
	if(drawsurf.x+dw>=viewx+vieww) dw=viewx+vieww-drawsurf.x;
	
	dh=u->clipy2-posy+viewy-drawsurf.y+1;
	if(dh<=0) return;
	if(drawsurf.y+dh>=viewy+viewh) dh=viewy+viewh-drawsurf.y;

	
	Parent->drawsurf.x=drawsurf.x-viewx+posx;
	Parent->drawsurf.y=drawsurf.y-viewy+posy;
	
	ggl_bitblt(&(Parent->drawsurf),&drawsurf,dw,dh);
	
}

// DRAWING FUNCTIONS
void gListBox::RedrawFrame()
{
	// DRAW NON-CLIENT ITEMS DIRECTLY INTO PARENT SURFACE
		gControl *P=Parent;
		int x,y,w,h;
		
		x=ncx;
		y=ncy;
		w=ncwidth;
		h=ncheight;
		if(StyleFlags&S_HASBORDER) {

			if(StyleFlags&S_FOCUSED) {
			P->HLine(y,x,x+w-1,GGRAY12);
			P->HLine(y+h-1,x+1,x+w-1,GGRAY5);
			P->VLine(x,y+1,y+h-1,GGRAY12);
			P->VLine(x+w-1,y+1,y+h-1,GGRAY5);

			}
			else {
				P->HLine(y,x,x+w-1,GGRAY6);
				P->HLine(y+h-1,x+1,x+w-1,GGRAY4);
				P->VLine(x,y+1,y+h-1,GGRAY6);
				P->VLine(x+w-1,y+1,y+h-1,GGRAY4);
			}
			// CLEAN 1 PIXEL SEPARATION BETWEEN BORDER AND TEXT
			P->HLine(y+1,x+1,x+w-2,BkColor);
			P->HLine(y+h-2,x+1,x+w-2,BkColor);
			P->VLine(x+1,y+2,y+h-2,BkColor);
			P->VLine(x+w-2,y+2,y+h-2,BkColor);
			
			if(y+h-2>posy+viewh) {
				// CLEAN RECTANGLE UNDER TEXT
				P->Rect(x+2,posy+viewh,x+w-2,y+h-2,BkColor);
			}

		}
		else {
			if(y+h>posy+viewh) {
				// CLEAN RECTANGLE UNDER TEXT
				P->Rect(x,posy+viewh,x+w-1,y+h-1,BkColor);
			}
		}
		// DRAW SCROLL BARS
		if(scrollh) DrawHScrollBar();
		if(scrollv) DrawVScrollBar();

}

void gListBox::DrawVScrollBar()
{
	gControl *P=Parent;

	int x,y,clen,cpos,h;

	x=ncx+ncwidth-SCROLLBARWIDTH;
	y=ncy;
	h=ncheight;
	if(StyleFlags&S_HASBORDER) { ++y; --x; h-=2*BORDERWIDTH; }
	
	// DRAW FRAME
	P->VLine(x,y,y+h-1,GGRAY7);
	P->Rect(x+1,y,x+SCROLLBARWIDTH-2,y+h-1,GGRAY3);
	P->HLine(y,x+1,x+SCROLLBARWIDTH-1,GGRAY7);
	P->VLine(x+SCROLLBARWIDTH-1,y,y+h-1,GWHITE);
	P->HLine(y+h-1,x+1,x+SCROLLBARWIDTH-1,GWHITE);

	// DRAW CURSOR
	if(NumItems>0) clen=((h-3)*(DispItems))/NumItems;
	else return;
	if(clen<1) clen=1;
	if(clen>h-3) clen=h-3;
	cpos=(ViewItem*(h-3-clen))/((NumItems-DispItems));

	P->Rect(x+1,y+cpos+1,x+SCROLLBARWIDTH-2,y+cpos+clen+1,GBLACK);

}

void gListBox::DrawHScrollBar()
{
	gControl *P=Parent;

	int x,y,clen,cpos;
	int ww;

	x=posx;
	y=posy+viewh;
	ww=x+vieww-1;
	if(StyleFlags&S_HASBORDER) { x--; y++; ww+=2; }	// MOVE SCROLLBAR TOWARDS EDGE
	
	// DRAW FRAME
	P->HLine(y,x,ww,GGRAY7);
	P->Rect(x,y+1,ww,y+SCROLLBARWIDTH-2,GGRAY3);
	P->VLine(x,y+1,y+SCROLLBARWIDTH-1,GGRAY7);
	P->HLine(y+SCROLLBARWIDTH-1,x,ww,GWHITE);
	P->VLine(ww,y+1,y+SCROLLBARWIDTH-1,GWHITE);

	// DRAW CURSOR
	ww=vieww;
	if(StyleFlags&S_HASBORDER) ww+=2;
	if(sizex>0) {
		clen=((ww-2)*(ww-2))/sizex;
	}
	else return;
	if(clen<1) clen=1;
	cpos=(viewx*ww)/sizex;
	
	P->Rect(x+cpos+1,y+1,x+cpos+clen+1,y+SCROLLBARWIDTH-2,GBLACK);



}





BOOL gListBox::DefaultEvent(gEvent *e)
{
	gUpdate u;
	gEvent ep;
	
	if(e->Type==GEVENT_KEYB) {

		// IMPLEMENT BASIC KEYBOARD HANDLING
		switch(e->Arg)
		{
		// ADD SPECIAL KEYS HERE
		case KM_PRESS | KB_LF:
		SetViewX(viewx-4);
		return TRUE;
		case KM_PRESS | KB_RT:
		SetViewX(viewx+4);
		return TRUE;
		case KM_PRESS | KB_UP:
		if(SelectedItem>0) {
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_BEFORECHANGE;
		ep.Data=SelectedItem;
		if(DoEventExclusive(&ep)) {
			if(!ep.Data) return TRUE;
		}
				Unselect(SelectedItem);
				MoveSelectionUp(1);
				Select(SelectedItem);
				ep.Type=GEVENT_GUI;
				ep.Arg=GEVGUI_AFTERCHANGE;
				ep.Data=SelectedItem;
				DoEventExclusive(&ep);
		}
		return TRUE;
		case KM_PRESS | KB_DN:
		if(SelectedItem<NumItems-1) {
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=SelectedItem;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}
		Unselect(SelectedItem);
		MoveSelectionDown(1);
		Select(SelectedItem);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);
		}
		return TRUE;
		case KM_PRESS | KB_DN | SHIFT_RS | SHIFT_RSHOLD:
		if(SelectedItem<NumItems-1) {
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=SelectedItem;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}

			Unselect(SelectedItem);
		MoveSelectionDown(DispItems);
		Select(SelectedItem);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);

		}
		return TRUE;
		case KM_PRESS | KB_UP | SHIFT_RS | SHIFT_RSHOLD:
		if(SelectedItem>0) {
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=SelectedItem;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}

			Unselect(SelectedItem);
		MoveSelectionUp(DispItems);
		Select(SelectedItem);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);

		}
		return TRUE;
		case KM_PRESS | KB_UP | SHIFT_RS:
		if(SelectedItem>0) {
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=SelectedItem;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}

			Unselect(SelectedItem);
		MoveSelectionUp(SelectedItem);
		Select(SelectedItem);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);

		}
		return TRUE;
		case KM_PRESS | KB_DN | SHIFT_RS:
		if(SelectedItem<NumItems-1) {
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=SelectedItem;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}

			Unselect(SelectedItem);
		MoveSelectionDown(NumItems-1-SelectedItem);
		Select(SelectedItem);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);

		}
		return TRUE;
		case KM_PRESS | KB_LF | SHIFT_RS:
		SetViewX(0);
		return TRUE;
		case KM_PRESS | KB_RT | SHIFT_RS:
		SetViewX(sizex-vieww);
		return TRUE;
		case KM_PRESS | KB_W:
		case KM_PRESS | KB_SPC:
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_BEFORECHECK;
		ep.Data=SelectedItem;
		if(DoEventExclusive(&ep)) {
			if(!ep.Data) return TRUE;
		}	
		SetCheck(SelectedItem,!IsChecked(SelectedItem));
		ClearClipRegion();
		RedrawAll();
		Invalidate();
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHECK;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);
		
		return TRUE;
		case KM_PRESS | KB_ADD:
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_BEFOREEXPAND;
		ep.Data=SelectedItem;
		if(DoEventExclusive(&ep)) {
			if(!ep.Data) return TRUE;
		}	
		ClearClipRegion();
		Expand(SelectedItem);
		RedrawAll();
		Invalidate();
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTEREXPAND;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);
		return TRUE;
		case KM_PRESS | KB_SUB:
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_BEFORECOLLAPSE;
		ep.Data=SelectedItem;
		if(DoEventExclusive(&ep)) {
			if(!ep.Data) return TRUE;
		}	
		ClearClipRegion();
		Collapse(SelectedItem);
		RedrawAll();
		Invalidate();
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCOLLAPSE;
		ep.Data=SelectedItem;
		DoEventExclusive(&ep);		
		return TRUE;
		
		
		// USE DEFAULT KEY BEHAVIOR
		default:
		return FALSE;

		}

	}
	
	if(e->Type==GEVENT_GUI) {
		switch(e->Arg)
		{
		case GEVGUI_GETFOCUS:
			// SELECTION HILIGHT
			Select(SelectedItem);

			StyleFlags|=S_FOCUSED;
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth-1;
			u.clipy2=ncy+ncheight-1;
			Parent->Invalidate(&u,this);
			return TRUE;
		case GEVGUI_LOSEFOCUS:
			// SELECTION HILIGHT
			Unselect(SelectedItem);
			StyleFlags&=~S_FOCUSED;
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth-1;
			u.clipy2=ncy+ncheight-1;
			Parent->Invalidate(&u,this);
			return TRUE;
		}
		
	}


	return FALSE;

	
}

void gListBox::Select(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	if(!(Items[Index].StateFlags&LS_SELECTED)) {

		Items[Index].StateFlags|=LS_SELECTED;
		RedrawSingleItem(Index);
		return;
	}
}

void gListBox::Unselect(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	if((Items[Index].StateFlags&LS_SELECTED)) {
		

		Items[Index].StateFlags&=~LS_SELECTED;
		RedrawSingleItem(Index);
		return;
	}
}


void gListBox::Hide(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	gListItem *i=Items+Index;
	int baselevel=i->TreeLevel;
	i->StateFlags|=LS_HIDDEN;
	

	
	while(Index<NumItems-1) {
		++i;
		++Index;
		if(i->TreeLevel<=baselevel) break;
	if(i->TreeLevel==baselevel+1) Hide(Index);
	}
	return;
}


void gListBox::Unhide(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	gListItem *i=Items+Index;
	int baselevel=i->TreeLevel;
	i->StateFlags&=~LS_HIDDEN;

	if( (i->StateFlags&LS_EXPANDABLE)&& (!(i->StateFlags&LS_EXPANDED))) return; 

	while(Index<NumItems-1) {
		++i;
		++Index;
		if(i->TreeLevel<=baselevel) break;
	if(i->TreeLevel==baselevel+1) Unhide(Index);
	}
	return;
}
void gListBox::Expand(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	gListItem *i=Items+Index;
	if(!(i->StateFlags&LS_EXPANDABLE)) return;
	i->StateFlags|=LS_EXPANDED;
    Unhide(Index);
	
}
void gListBox::Collapse(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	
	gListItem *i=Items+Index;
	if(!(i->StateFlags&LS_EXPANDABLE)) return;
    Hide(Index);
	i->StateFlags&=~LS_EXPANDED;
	i->StateFlags&=~LS_HIDDEN;
}


BOOL gListBox::HasChildren(int Index)
{
	if(Index<0 || Index>=NumItems-1) return FALSE;
	if(Items[Index+1].TreeLevel==Items[Index].TreeLevel+1) return TRUE;
	return FALSE;
}
BOOL gListBox::IsExpanded(int Index)
{
	if(Index<0 || Index>=NumItems) return FALSE;
	if( (Items[Index].StateFlags&(LS_EXPANDED|LS_EXPANDABLE))==(LS_EXPANDED|LS_EXPANDABLE))
		return TRUE;
	return FALSE;
}

void gListBox::SetCheck(int Index,BOOL value)
{
	if(Index<0 || Index>=NumItems) return;
	
	gListItem *i=Items+Index;
	int orgindex=Index;
	
	if(!(i->StateFlags&LS_HASCHECK)) return;
	
	// UPDATE ALL CHILDREN
	int baselevel=i->TreeLevel;
	if(value) i->StateFlags|=LS_CHECKED;
	else i->StateFlags&=~LS_CHECKED;
	i->StateFlags&=~LS_UNKNOWN;
	
	while(Index<NumItems-1) {
		++i;
		++Index;
		if(i->TreeLevel<=baselevel) break;
	if(i->StateFlags&LS_HASCHECK) {
		if(value) i->StateFlags|=LS_CHECKED;
		else i->StateFlags&=~LS_CHECKED;
		i->StateFlags&=~LS_UNKNOWN;

	}
	}
	
	// SET PARENT STATE TO UNKNOWN/PARTIAL
	for(--orgindex,i=Items+orgindex;orgindex>=0;--orgindex,--i) {
		if(i->TreeLevel==baselevel-1) {
			if(i->StateFlags&LS_HASCHECK) {
				i->StateFlags|=LS_UNKNOWN;
			}
			--baselevel;
		}
	}
	
	return;
}
BOOL gListBox::IsChecked(int Index)
{
	if(Index<0 || Index>=NumItems) return FALSE;
	
	if((Items[Index].StateFlags&(LS_HASCHECK|LS_CHECKED|LS_UNKNOWN))==(LS_HASCHECK|LS_CHECKED)) return TRUE;
	return FALSE;
}










BOOL gListBox::MoveSelectionUp(int n)
{
	int newsel=SelectedItem,prevsel=SelectedItem;
	for(;n>0;--n) {
		do {
			--newsel;
			if(newsel<0) break;
		} while(Items[newsel].StateFlags&LS_HIDDEN);
		
		if(newsel<0) break;
		prevsel=newsel;
	}
	if( (newsel<0) && (prevsel==SelectedItem) ) return FALSE;
	if(newsel<0) newsel=prevsel;
	SelectedItem=newsel;
	
	EnsureVisible(SelectedItem);
	return TRUE;
}

BOOL gListBox::MoveSelectionDown(int n)
{
	int newsel=SelectedItem,prevsel=SelectedItem;
	for(;n>0;--n) {
		do {
			++newsel;
			if(newsel>=NumItems) break;
		} while(Items[newsel].StateFlags&LS_HIDDEN);
		
		if(newsel>=NumItems) break;
		prevsel=newsel;
	}
	if( (newsel>=NumItems) && (prevsel==SelectedItem) ) return FALSE;
	if(newsel>=NumItems) newsel=prevsel;
	SelectedItem=newsel;
	
	EnsureVisible(SelectedItem);
	return TRUE;
}



void gListBox::RedrawSingleItem(int Index)
{
	int f,nitem,y;
	if(Index<ViewItem) return;
	for(f=0,nitem=ViewItem,y=0;f<DispItems;++f)
	{
		if(nitem==Index) {
			DrawItem(nitem,y);
			gUpdate u;
			u.clipx=viewx;
			u.clipx2=viewx+vieww;
			u.clipy=y;
			u.clipy2=y+ItemHeight;
			Invalidate(&u,this);
		}
		do {
		++nitem;
		if(nitem>=NumItems) return;
		} while(Items[nitem].StateFlags&LS_HIDDEN);
		y+=ItemHeight;
	}	
}

void gListBox::InvalidateVScrollBar()
{
	int x,y,h;
	x=ncx+ncwidth-SCROLLBARWIDTH;
	y=ncy;
	h=ncheight;
	if(StyleFlags&S_HASBORDER) { ++y; --x; h-=2*BORDERWIDTH; }
	
	gUpdate u;
	u.clipx=x;
	u.clipx2=x+SCROLLBARWIDTH-1;
	u.clipy=y;
	u.clipy2=y+h-1;
	Parent->Invalidate(&u,this);
}



void gListBox::EnsureVisible(int Index)
{
	int scrollcount,i;
	
if(Index<0 || Index>=NumItems) return;

if(Index<ViewItem) {
	// SCROLL DOWN
	
	// CALCULATE NUMBER OF LINES TO SCROLL DOWN
	ClearClipRegion();

	int newi;
	for(i=ViewItem,scrollcount=0;i>Index;++scrollcount) {
		newi=PrevVisibleItem(i);
		if(newi<0) break;
		i=newi;
	}
	
	ViewItem=i;

	if(scrollcount<DispItems) {
		drawsurf.x=0;
		drawsurf.y=0;
		ggl_scrolldn(&drawsurf,sizex,DispItems*ItemHeight,scrollcount*ItemHeight);
		int f,nitem,y;
		for(f=0,nitem=ViewItem,y=0;f<scrollcount;++f)
		{
			while(!DrawItem(nitem,y)) { ++nitem; }
			++nitem;
			y+=ItemHeight;
		}
		
		if(scrollv) InvalidateVScrollBar();
		Invalidate();
		return;
	}
	
	RedrawAll();
	if(scrollv) InvalidateVScrollBar();
	Invalidate();
	return;
}

// FIND LAST VISIBLE ITEM

int nitem,nitem2;
for(i=0,nitem=ViewItem;i<DispItems-1;++i)
{
	nitem2=NextVisibleItem(nitem);
	if(nitem2<0) break;
	nitem=nitem2;
}
if(Index<=nitem) return;

// CALCULATE SCROLL COUNT
ClearClipRegion();

for(i=nitem,scrollcount=0;i<Index;++scrollcount) {
	nitem2=NextVisibleItem(i);
	if(nitem2<0) break;
	i=nitem2;
}

if(scrollcount<DispItems) {
	drawsurf.x=0;
	drawsurf.y=0;
	ggl_scrollup(&drawsurf,sizex,DispItems*ItemHeight,scrollcount*ItemHeight);
	int y;
	for(i=0,nitem=Index,y=(DispItems-1)*ItemHeight;i<DispItems-1;++i)
	{
		if(i<scrollcount) { while(!DrawItem(nitem,y)) { --nitem; } }
		nitem=PrevVisibleItem(nitem);
		y-=ItemHeight;
	}
	ViewItem=nitem;
	if(scrollv) InvalidateVScrollBar();
	Invalidate();
	return;
}

for(i=0,nitem=Index;i<DispItems-1;++i)
{
	nitem2=PrevVisibleItem(nitem);
	if(nitem2<0) break;
	nitem=nitem2;
}
ViewItem=nitem;

RedrawAll();
if(scrollv) InvalidateVScrollBar();
Invalidate();
return;

}

void gListBox::SetSelection(int Index)
{
	gEvent ep;
	
	if(Index<0 || Index>=NumItems) return;
	ep.Type=GEVENT_GUI;
	ep.Arg=GEVGUI_BEFORECHANGE;
	ep.Data=SelectedItem;
	if(DoEventExclusive(&ep)) {
		if(!ep.Data) return;
	}
			Unselect(SelectedItem);
			
			if(Items[Index].StateFlags&LS_HIDDEN) {
				// DON'T SELECT A HIDDEN ITEM
				int i=PrevVisibleItem(Index);
				if(i<0) i=NextVisibleItem(Index);
				if(i>=0) SelectedItem=i;
			}
			else SelectedItem=Index;
			Select(SelectedItem);
			EnsureVisible(SelectedItem);
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_AFTERCHANGE;
			ep.Data=SelectedItem;
			DoEventExclusive(&ep);

}
int gListBox::NextItem(int Index)
{
	int baselevel=Items[Index].TreeLevel;
	if(Index<0) return -1;
	while(Index<NumItems-1) {
		++Index;
		if(Items[Index].TreeLevel<baselevel) return -1;
		if(Items[Index].TreeLevel==baselevel) return Index;
	}
	return -1;
}

int gListBox::NextVisibleItem(int Index)
{
	if(Index<0) return -1;
	while(Index<NumItems-1) {
		++Index;
		if(!(Items[Index].StateFlags&LS_HIDDEN)) return Index;
	}
	return -1;
}

int gListBox::PrevVisibleItem(int Index)
{
	if(Index>=NumItems) return -1;
	while(Index>0) {
		--Index;
		if(!(Items[Index].StateFlags&LS_HIDDEN)) return Index;
	}
	return -1;
}


int gListBox::PrevItem(int Index)
{
	int baselevel=Items[Index].TreeLevel;
	if(Index>=NumItems) return -1;
	while(Index>0) {
		--Index;
		if(Items[Index].TreeLevel<baselevel) return -1;
		if(Items[Index].TreeLevel==baselevel) return Index;
	}
	return -1;
	
}
int gListBox::ChildItem(int Index)
{
	if(Index<0 || Index>=NumItems-1) return -1;
	int baselevel=Items[Index].TreeLevel;
	if(Items[Index+1].TreeLevel==baselevel+1) return Index+1;
	return -1;
}

int gListBox::ParentItem(int Index)
{
	int baselevel=Items[Index].TreeLevel-1;
	if(Index>=NumItems) return -1;
	while(Index>0) {
		--Index;
		if(Items[Index].TreeLevel<baselevel) return -1;
		if(Items[Index].TreeLevel==baselevel) return Index;
	}
	return -1;
	
}

char *gListBox::FullPath(int BaseIndex,int Index,const char *Separator)
{
	if(Index>=NumItems || Index<0) return NULL;

	int seplen=strlen(Separator),len;
	int parent=Index,totallength=0;
	// DETERMINE PATH LENGTH
	do {
		totallength+=strlen(Items[parent].Text);
		parent=ParentItem(parent);
		if(parent==BaseIndex) break;
		totallength+=seplen;
	} while(parent>=0);
	
	char *buffer=new char[totallength+1];
	if(!buffer) return NULL;
	
	parent=Index;
	buffer[totallength]=0;
	// DO THE ACTUAL COPY
	do {
		len=strlen(Items[parent].Text);
		totallength-=len;
		memcpy(buffer+totallength,Items[parent].Text,len);
		parent=ParentItem(parent);
		if(parent==BaseIndex) break;
		totallength-=seplen;
		memcpy(buffer+totallength,Separator,seplen);
	} while(parent>=0);
	return buffer;
}

void gListBox::ExpandAll(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	do {
		ExpandAll(ChildItem(Index));
		Expand(Index);
		Index=NextItem(Index);
	} while(Index>=0 && Index<NumItems);
	
	ClearClipRegion();
	RedrawAll();
	Invalidate();	
	return;
}

void gListBox::CollapseAll(int Index)
{
	if(Index<0 || Index>=NumItems) return;
	do {
		CollapseAll(ChildItem(Index));
		Collapse(Index);
		Index=NextItem(Index);
		
	} while(Index>=0 && Index<NumItems);
	
	ClearClipRegion();
	RedrawAll();
	Invalidate();	

	return;
	
}

