//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***


// $Id: vsprintf.c 393 2007-03-04 23:17:29Z ingo $

#include <hpgcc3.h>

char *_print_out_copy(char *t,char *s);


// This is the workhorse for all *printf variants 

/*
    TODO: 
        Approximate to C lib standard formats
        =====================================
        
        * Correct prefix interpretation u,l,ll
        * More formats (?)
        * Remove non standard formats b,U,L (superset (?))

*/

// REWRITTEN BY CLAUDIO - APR 2006


int
vsprintf(char *out, const char *fmt, va_list argp)
{
const char *p,*orgout=out;
struct conversion_data data;
unsigned long long tempnum;
int modifier;
XREAL xr;

for(p = fmt; *p != '\0'; p++) {
    
    if(*p != '%') {
        *out++ = *p;
        continue;
    }
    
	memset((void *)&data,0,sizeof(struct conversion_data));
	data.precision=-1;

	
	// GET FORMAT FLAGS FIRST
	
	do {
    p++;

	if(*p=='-') { data.flags|=1; continue; }
	if(*p=='+') { data.flags|=2; continue; }
	if(*p==' ') { data.flags|=8; continue; }
	if(*p=='0') { data.flags|=4; continue; }
	if(*p=='#') { data.flags|=48; continue; }
	break;
	}
	while(1);
	
	// GET WIDTH SPECIFIER
    if(isdigit(*p) )
	{
	tempnum=0;
	p=_str2ull(p,&tempnum,10);
	data.width=(int)tempnum;
    }
    
	// GET PRECISION SPECIFIER
	if(*p=='.') {
	tempnum=0;
	p=_str2ull(++p,&tempnum,10);
	data.precision=(int)tempnum;
	}
	

	// MODIFIER CHARACTERS
	modifier=0;
    if(*p=='h') { modifier=1; ++p; }
	if(*p=='l') { modifier=2; ++p; }
	if(*p=='l') { modifier++; ++p; }
	

	// FINAL FORMAT
	      
    if (*p ==  's') {
		char *s;
        if (!(s = va_arg(argp, char *)))
            s = "<NULL>";
		int padding=data.width-strlen(s);
		if(data.flags&1) {
		out = _print_out_copy(out,s);
		while(padding-->0) *out++=' ';
		} else {
		while(padding-->0) *out++=' ';
		out = _print_out_copy(out,s);
		}
         continue;
    }

    if (*p == 'c') {
        *out++=(char)va_arg(argp, int);
        continue;
    }

    if (*p == '%') {
        *out++='%';
        continue;
    }

	
	if(!strchr("oxXudifgGeE",*p)) continue;


	// NUMERIC TYPES FROM HERE ON
	
    if (strchr("oxXpbudi",*p)) {
	// ALL INTEGER TYPES
	data.precision=0;
	data.base=10;
	data.type=2;
	data.flags&=31;		// REMOVE FORCED DECIMAL POINT
	if(*p=='b') data.base=2;
	if(*p=='o') data.base=8;
	if(*p=='x') data.base=16;
	if(*p=='p') { data.base=16; data.width=8; data.flags|=4; }
	if(*p=='X') { data.base=16; data.flags|=64; }
	if(modifier==3) {	// LONG LONG
	tempnum=(unsigned long long)va_arg(argp, unsigned long long);
	} else 	tempnum=(unsigned long long)va_arg(argp, unsigned int);

	if(*p=='i' || *p=='d') {
	// SIGN EXTEND 32-BIT VALUES
		if(modifier!=3) {  if(tempnum&0x80000000LL) tempnum|=0xffffffff00000000LL; }
	data.type=1;
	xr_llong2xr(tempnum,&xr);
	} else xr_ullong2xr(tempnum,&xr);
	
	}
	else {
	// FLOATING POINT NUMBERS
	data.base=10;
	if(data.precision<0) data.precision=6; // default to 6 figures
	if(*p=='E' || *p=='G') data.flags|=64;
	data.type=((*p&223)-'D')<<2;
	
	if(modifier==3) {		// %llf = extended real
	memcpy(&xr,(va_arg(argp, XREAL *)),sizeof(XREAL));
	} else {
	double tempdbl;
	tempdbl=va_arg(argp, double);
	xr_double2xr(tempdbl,&xr);
	}
	
	}

	// FINALLY PRINT
	
	if(data.base==10) xr_base10(&xr,&xr);

	char *newstring=xr_xrtoa(&xr,&data);
	
    out = _print_out_copy(out,newstring);
	free(newstring);    
    }


    *out = '\0';
    return (int)(out-orgout);

}




/*
int
vsprintf(char *out, const char *fmt, va_list argp)
{
const char *p;

char fmtbuf[256];
char padbuf[256];
char hexbuf[16];
char *s=NULL;

int lead_zero, pad, minus, entry, fpdigits, fpscale;

for(p = fmt; *p != '\0'; p++) {
    
    if(*p != '%') {
        *out++ = *p;
        continue;
    }
    
    lead_zero = pad = minus = fpdigits = fpscale = 0;
    entry = 1;
    
    p++;
    
    if ((minus = *p == '-'))
        p++;
    
    while(isdigit(*p) ) {
        if (entry) {
            lead_zero = (*p == '0');
            entry = 0;
        }
        pad = 10 * pad + ( *p & 0xF );
        p++;
    }
    
    if ((fpscale = *p == '.')) {
        p++;
        while(isdigit(*p) ) {
            fpdigits = 10 * fpdigits + ( *p & 0xF );
            p++;
        }
    }
        
        
    
        
    if (*p ==  's') {
        if (!(s = va_arg(argp, char *)))
            s = "<NULL>";
        out = _print_out_copy(out,pad ? xpad(padbuf,s,pad,minus,' ') : s);
        
        goto t_break;
    }
    
    if (*p == 'd' || *p == 'i') {
    
        s = itoa(va_arg(argp, int), fmtbuf, 10);
        goto numeric_puts;
    }
    
    if (*p == 'c') {
        *out++=(char)va_arg(argp, int);
        goto t_break;
    }
    
    if (*p == 'f' || *p == 'g' || *p == 'e' || *p == 'E') {
    
        int old_fpdigits = get_decimal_digits();
        
        if (fpscale)
            set_decimal_digits(fpdigits);
        s = __dtoa(dround(va_arg(argp,double),fpscale ? fpdigits : old_fpdigits), fmtbuf,*p);
        if (fpscale)
            set_decimal_digits(old_fpdigits);
        goto numeric_puts;
    }
    
    if (*p == 'x' || *p == 'X') {
        s = utoa(va_arg(argp, unsigned), fmtbuf, 16);
        if (*p == 'X')
            s = upper(hexbuf,fmtbuf);
        goto numeric_puts;
    }
    
    if (*p == 'p') {
        s = utoa((unsigned) va_arg(argp, void *), fmtbuf, 16);
        goto numeric_puts;
    }
    
    
    //The 'b' format has been added for debug only, this is NOT standard C format
    if (*p == 'u' || *p == 'b') {
    
        s = utoa(va_arg(argp, unsigned), fmtbuf, *p == 'b' ? 2 : 10);
        //puts(pad ? xpad(padbuf,s,pad,!minus,lead_zero ? '0' : ' ') : s);
        goto numeric_puts;
    }
        
    
    if (*p == 'U') {
        
        s = ulltoa(va_arg(argp, unsigned long long), fmtbuf,10);
        goto numeric_puts;
    }
        
    if (*p ==  'L') {
        s = lltoa(va_arg(argp, long long), fmtbuf,10);
        goto numeric_puts;
    }
    

    if (*p == '%') {
        *out++='%';
        goto t_break;
    }
    
        
numeric_puts:
    out = _print_out_copy(out,pad ? xpad(padbuf, s, pad, ! minus, lead_zero ? '0' : ' ') : s);
    // goto t_break;
        

t_break:    ;
        
    
    }


    *out = '\0';
    return strlen(out);

}
*/


