//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



/* $Id: malloc.c 599 2008-09-01 16:37:15Z claudio $ */


#include <stdlib.h>
#include <master_globals.h>

// ROMLIB CHECKED



#ifndef ENOMEM
#define ENOMEM -90012
#endif

#define CAST_INT    int
#define BRKSIZE        4096

#define PTRSIZE        ((int) sizeof(void *))
#ifndef ALIGN
#define ALIGN(x,a)    (((x) + (a - 1)) & ~(a - 1))
#endif
#define NEXT_SLOT(p)    (* (void **) ((p) - PTRSIZE))
#define NEXT_FREE(p)    (* (void **) (p))


extern void *_sbrk(int);
extern int _brk(void *);



void __free(void *ptr);
BOOL __isusedheapptr(void *ptr);




static int grow(size_t len)
{
  register char *p;

  __mg->errno = ENOMEM;
  if ((char *) __mg->_malloc_top + len < (char *) __mg->_malloc_top
      || (p = (char *)ALIGN((CAST_INT)__mg->_malloc_top + len, BRKSIZE)) < (char *) __mg->_malloc_top 
      || _brk(p) != 0)
    return(0);
  NEXT_SLOT((char *)__mg->_malloc_top) = p;
  NEXT_SLOT(p) = 0;
  __free(__mg->_malloc_top);
  __mg->_malloc_top = p;
  return 1;
}


void *
malloc(size_t size)
{
  register char *prev, *p, *next, *new;
  register unsigned len, ntries;

  if (size == 0) return NULL;
  __mg->errno = ENOMEM;
  for (ntries = 0; ntries < 2; ntries++) {
    if ((len = ALIGN(size, PTRSIZE) + PTRSIZE) < 2 * PTRSIZE)
        return NULL;
    if (__mg->_malloc_bottom == 0) {
        if ((p = _sbrk(2 * PTRSIZE)) == (char *) -1)
            return NULL;
        p = (char *) ALIGN((CAST_INT)p, PTRSIZE);
        p += PTRSIZE;
        __mg->_malloc_top = __mg->_malloc_bottom = p;
        NEXT_SLOT(p) = 0;
    }

    for (prev = 0, p = __mg->_malloc_empty; p != 0; prev = p, p = NEXT_FREE(p)) {
        next = NEXT_SLOT(p);
        new = p + len;    
        if (new > next || new <= p)
            continue;        
        if (new + PTRSIZE < next) {    
            NEXT_SLOT(new) = next;
            NEXT_SLOT(p) = new;
            NEXT_FREE(new) = NEXT_FREE(p);
            NEXT_FREE(p) = new;
        }
        if (prev)
            NEXT_FREE(prev) = NEXT_FREE(p);
        else
            __mg->_malloc_empty = NEXT_FREE(p);

        return p;
    }
    if (grow(len) == 0)
        break;
  }

  return NULL;
}


void
__free(void *ptr)
{
  register char *prev, *next;
  char *p = ptr;

  if(!p) return;


  for (prev = 0, next = __mg->_malloc_empty; next != 0; prev = next, next = NEXT_FREE(next))
    if (p < next)
        break;
  NEXT_FREE(p) = next;
  if (prev)
    NEXT_FREE(prev) = p;
  else
    __mg->_malloc_empty = p;
  if (next) {
    if (NEXT_SLOT(p) == next) {        
        NEXT_SLOT(p) = NEXT_SLOT(next);
        NEXT_FREE(p) = NEXT_FREE(next);
    }
  }
  if (prev) {
    if (NEXT_SLOT(prev) == p) {        
        NEXT_SLOT(prev) = NEXT_SLOT(p);
        NEXT_FREE(prev) = NEXT_FREE(p);
    }
  }
}

// IMPROVED HEAP SAFETY FOR FINAL USERS - CLAUDIO 12-02-2006

void
free(void *ptr)
{
	if(__isusedheapptr(ptr)) __free(ptr);
  
}





// ADDED BY CLAUDIO - 12/02/2006
// DETECT IF A POINTER FALLS WITHIN A BLOCK
// RETURNS ADDRESS AND SIZE OF THE CONTAINER BLOCK, AND IF IT'S A FREE BLOCK OR USED BLOCK


int getheapblockinfo(void *ptr, void **blkstart,size_t *blksize)
{
  register char *next,*free;
  char *p = ptr;

  if (p<(char *)__mg->_malloc_bottom || p>=((char *)__mg->_malloc_top)-PTRSIZE) return INVALID_PTR;

  for (next = __mg->_malloc_bottom,free=__mg->_malloc_empty; next != 0; next = NEXT_SLOT(next)) {

	while(free && next>free) free=NEXT_FREE(free); // ALWAYS POINT TO THE FREE BLOCK ABOVE next
	if(p>=next && p< (char *)(NEXT_SLOT(next)-PTRSIZE)) {
	// FOUND BLOCK
	*blkstart=next;
	*blksize=(size_t) ( ( ((char *)NEXT_SLOT(next))-next)-PTRSIZE);
	return (free==next)? BLOCK_FREE:BLOCK_USED;
	
	}
  
  
  }

	// SHOULDN'T HAPPEN UNLESS HEAP IS CORRUPT OR POINTER IS POINTING TO THE NEXT_SLOT() AREA
	// IN WHICH CASE IT IS AN INVALID POINTER
	return INVALID_PTR;

}

BOOL __isusedheapptr(void *ptr)
{
  void *blkstart;
  size_t blksize; 

// IMPROVED FINAL-USER SAFETY CHECK BY CLAUDIO 12-2-2006
  if(getheapblockinfo(ptr,&blkstart,&blksize)!=BLOCK_USED) return FALSE;	// RETURN IF NOT A USED BLOCK
  if(ptr != blkstart) return FALSE; // RETURN IF NOT POINTING EXACTLY TO A VALID HEAP BLOCK
  
 return TRUE;
  
}

// RETURN FREE MEMORY
int freemem()
{
	register unsigned int stack_ptr asm ("sp");
	unsigned int base;
	
    base=(unsigned int) ( (__mg->__heap_ptr == 0)? (unsigned int)(__mg->_heap_base_addr):(unsigned int)(__mg->__heap_ptr));
 
    return (int) (stack_ptr-base);
}
