//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: __launch_poke.c 393 2007-03-04 23:17:29Z ingo $

#include <hpgcc3.h>

// helper functions: standard versions use globals, so they won't work here


// sat_addr CAN BE: 1) SATURN VISIBLE ADDRESS OR 2) (BANK&0xff)<<20 + ADDRESS
// WHERE BANK= 0-15 --> FLASH, -1 = MAIN RAM, -2=ERAM AS RETURNED BY __launch_getbanknum()

unsigned int
__launch_s2a(unsigned int sat_addr,SAT_CPU *cpu)
{

    if(sat_addr&0x0ff00000) { 		// ADDRESS HAS EMBEDDED BANK INFORMATION?
						if(sat_addr&0X0F000000) // BANK IS IN RAM?
							return (((sat_addr&0xfffff)>>1)+0x08000000);	// FOR RAM: ARM=SATURN>>1 + 0X08000000
						else
							return ((sat_addr&0xfffff)>>1)+((sat_addr&0x0ff00000)>>3)+0x20000;  // FOR FLASH: ARM=SATURN>>1 + (BANK+1)*0x20000
						}

	// NO BANK INFORMATION, USE VISIBLE SATURN ADDRESSES
    return cpu->Read[sat_addr >> 12] + (sat_addr & 0xfff)/2;
}

unsigned
__launch_peek(int sataddr,int nibbles,SAT_CPU *cpu)
{
    
    // GENERAL PEEK ROUTINE, UP TO 8 NIBBLES

    LONGLONG *ptr=(LONGLONG *)(__launch_s2a(sataddr,cpu)&0xfffffffc);	
	
    unsigned mask;
    int rot=(sataddr&7)<<2;
    
    mask=1<<(nibbles<<2);
    mask--;

return (unsigned)( (unsigned int)((*ptr)>>rot))&mask;
}



void __launch_poke(int sataddr,unsigned val,int nibbles,SAT_CPU *cpu)
{
    
    // GENERAL POKE ROUTINE, UP TO 8 NIBBLES
    

    LONGLONG *ptr=(LONGLONG *)(__launch_s2a(sataddr,cpu)&0xfffffffc);

    ULONGLONG lval,mask2;
    unsigned int mask;
    int rot=(sataddr&7)<<2;
    
    lval=*ptr;
    mask=1<<(nibbles<<2);
    mask--;
    val&=mask;
    mask2=((ULONGLONG)mask)<<rot;
    mask2=-mask2-1;
    *ptr=(((ULONGLONG)val)<<rot)|(lval&mask2);

}

unsigned
__launch_peekarm(int armaddr,int nibbles,SAT_CPU *cpu)
{
    
    // GENERAL PEEK ROUTINE, UP TO 8 NIBBLES

    LONGLONG *ptr=(LONGLONG *)((armaddr>>1)&0xfffffffc);

    unsigned mask;
    int rot=(armaddr&7)<<2;
    
    mask=1<<(nibbles<<2);
    mask--;

return (unsigned)( (unsigned int)((*ptr)>>rot))&mask;
}




void __launch_memmove(int satdest,int satsrc,int nibbles,SAT_CPU *cpu)
{
    
    // SATURN MOVEUP/MOVEDOWN IMPLEMENTATION
    
    int src=satsrc;
    int offset=satdest-satsrc,add;
	if(offset>0) {
	// MOVEUP
    src+=nibbles-8;
	add=-8;
	}
	else add=8;		// MOVEDOWN

    while(nibbles>=8) {
        __launch_poke(src+offset,__launch_peek(src,8,cpu),8,cpu);
        src+=add;
        nibbles-=8;
    }
    if(nibbles) {
		if(add<0) src+=8-nibbles;
		__launch_poke(src+offset,__launch_peek(src,nibbles,cpu),nibbles,cpu);
	}
}
